@@ -8,6 +8,7 @@ import yaml
 from pybedtools import BedTool
 import dask
 import re
+from typing import Union
 import warnings
 from .utilities import (
     highly_variable_methylation_feature,
@@ -24,7 +25,7 @@ class MCDS(xr.Dataset):
 
     __slots__ = ()
 
-    def __init__(self, dataset):
+    def __init__(self, dataset, obs_dim=None, var_dim=None):
         """
         Init MCDS
 
@@ -35,8 +36,67 @@ class MCDS(xr.Dataset):
         super().__init__(
             data_vars=dataset.data_vars, coords=dataset.coords, attrs=dataset.attrs
         )
+        self.obs_dim = obs_dim
+        self.var_dim = var_dim
         return
 
+    @property
+    def var_dim(self):
+        if self.attrs['var_dim'] == 'null':
+            return None
+        else:
+            return self.attrs['var_dim']
+
+    @var_dim.setter
+    def var_dim(self, var_dim):
+        if var_dim is not None:
+            if var_dim not in self.dims:
+                raise KeyError(
+                    f"{var_dim} does not occur in dimension names: {list(self.dims.keys())}"
+                )
+        else:
+            var_dim = 'null'
+        self.attrs['var_dim'] = var_dim
+        return
+
+    @property
+    def obs_dim(self):
+        if self.attrs['obs_dim'] == 'null':
+            return None
+        else:
+            return self.attrs['obs_dim']
+
+    @obs_dim.setter
+    def obs_dim(self, obs_dim):
+        if obs_dim is not None:
+            if obs_dim not in self.dims:
+                raise KeyError(
+                    f"{obs_dim} does not occur in dimension names: {list(self.dims.keys())}"
+                )
+        else:
+            obs_dim = 'null'
+        self.attrs['obs_dim'] = obs_dim
+        return
+
+    def _verify_dim(self, dim, mode):
+        if dim is None:
+            if mode == 'var':
+                if self.var_dim is None:
+                    raise ValueError('MCDS does not have var_dim set, please provide the name of var_dim')
+                else:
+                    return self.var_dim
+            else:
+                if self.obs_dim is None:
+                    raise ValueError('MCDS does not have obs_dim set, please provide the name of obs_dim')
+                else:
+                    return self.obs_dim
+        else:
+            if dim not in self.dims:
+                raise KeyError(
+                    f"{dim} does not occur in dimension names: {list(self.dims.keys())}"
+                )
+            return dim
+
     @classmethod
     def open(cls, mcds_paths, obs_dim="cell", use_obs=None, var_dim=None, split_large_chunks=True, engine=None):
         """
@@ -73,6 +133,20 @@ class MCDS(xr.Dataset):
             else:
                 _flat_mcds_paths.append(str(path))
 
+        # if var_dim is a list, open each separately and return merged MCDS:
+        if isinstance(var_dim, list):
+            ds_list = []
+            for _var_dim in var_dim:
+                ds = cls.open(_flat_mcds_paths,
+                              obs_dim=obs_dim,
+                              use_obs=use_obs,
+                              var_dim=_var_dim,
+                              split_large_chunks=split_large_chunks,
+                              engine=engine)
+                ds_list.append(ds)
+            ds = xr.merge(ds_list)
+            return cls(ds, obs_dim=obs_dim, var_dim=None).squeeze()
+
         # determine dataset var_dim
         _var_dims = set()
         _final_paths = []
@@ -91,8 +165,9 @@ class MCDS(xr.Dataset):
                                 _var_dims.add(_ds_var_dim)
                             else:
                                 raise ValueError(
-                                    f'MCDS paths containing multiple datasets {config["ds_region_dim"]}, '
-                                    'please specify which one to open with the "var_dim" parameter.'
+                                    f'MCDS paths containing multiple datasets '
+                                    f'{list(config["ds_region_dim"].keys())}, '
+                                    'please specify one or multiple names with the "var_dim" parameter.'
                                 )
                         else:
                             _ds_var_dim = config['region_dim']
@@ -107,8 +182,14 @@ class MCDS(xr.Dataset):
                 # single netcdf or zarr file
                 _final_paths.append(path)
 
-        if has_dataset and len(_var_dims) != 1:
-            raise ValueError(f'Some MCDS dataset has multiple var_dim, please specify var_dim parameter.')
+        if has_dataset:
+            if len(_var_dims) != 1:
+                raise ValueError(f'Some MCDS dataset has multiple var_dim, please specify var_dim parameter.')
+            else:
+                var_dim = _var_dims.pop()
+        for path in _final_paths:
+            if not pathlib.Path(path).exists():
+                raise FileNotFoundError(f'{path} does not exist.')
 
         # determine engine
         if engine is None:
@@ -132,15 +213,15 @@ class MCDS(xr.Dataset):
                     **{"array.slicing.split_large_chunks": split_large_chunks}):
                 use_obs_bool = ds.get_index(obs_dim).isin(use_obs)
                 ds = ds.sel({obs_dim: use_obs_bool})
-        return cls(ds).squeeze()
+        return cls(ds, obs_dim=obs_dim, var_dim=var_dim).squeeze()
 
     def add_mc_frac(
-        self,
-        var_dim,
-        da=None,
-        normalize_per_cell=True,
-        clip_norm_value=10,
-        da_suffix="frac",
+            self,
+            var_dim=None,
+            da=None,
+            normalize_per_cell=True,
+            clip_norm_value=10,
+            da_suffix="frac",
     ):
         """
         Add posterior mC rate data array for certain feature type (var_dim).
@@ -161,6 +242,8 @@ class MCDS(xr.Dataset):
         -------
 
         """
+        var_dim = self._verify_dim(dim=var_dim, mode='var')
+
         if da is None:
             da = f"{var_dim}_da"
         if da not in self.data_vars:
@@ -186,7 +269,7 @@ class MCDS(xr.Dataset):
         )
         self.add_mc_frac(*args, **kwargs)
 
-    def add_feature_cov_mean(self, var_dim, obs_dim="cell", plot=True):
+    def add_feature_cov_mean(self, obs_dim=None, var_dim=None, plot=True):
         """
         Add feature cov mean across obs_dim.
 
@@ -202,14 +285,14 @@ class MCDS(xr.Dataset):
         -------
         None
         """
-        feature_cov_mean = (
-            self[f"{var_dim}_da"]
-            .sel(count_type="cov")
-            .sum(dim="mc_type")
-            .mean(dim=obs_dim)
-            .squeeze()
-            .to_pandas()
-        )
+        obs_dim = self._verify_dim(obs_dim, mode='obs')
+        var_dim = self._verify_dim(var_dim, mode='var')
+
+        _da = self[f"{var_dim}_da"]
+        if 'mc_type' in _da.dims:
+            feature_cov_mean = _da.sel(count_type="cov").sum(dim="mc_type").mean(dim=obs_dim).squeeze().to_pandas()
+        else:
+            feature_cov_mean = _da.sel(count_type="cov").mean(dim=obs_dim).squeeze().to_pandas()
         self.coords[f"{var_dim}_cov_mean"] = feature_cov_mean
 
         print(
@@ -219,14 +302,15 @@ class MCDS(xr.Dataset):
             cutoff_vs_cell_remain(feature_cov_mean, name=f"{var_dim}_cov_mean")
         return
 
-    def add_cell_metadata(self, metadata, obs_name="cell"):
-        metadata.index.name = obs_name
-        mcds_index = self.get_index(obs_name)
+    def add_cell_metadata(self, metadata, obs_dim=None):
+        obs_dim = self._verify_dim(obs_dim, mode='obs')
+        metadata.index.name = obs_dim
+        mcds_index = self.get_index(obs_dim)
         for name, col in metadata.reindex(mcds_index).items():
-            self.coords[f"{obs_name}_{name}"] = col
+            self.coords[f"{obs_dim}_{name}"] = col
         return
 
-    def filter_feature_by_cov_mean(self, var_dim, min_cov=0, max_cov=999999):
+    def filter_feature_by_cov_mean(self, var_dim=None, min_cov=0, max_cov=999999):
         """
         filter MCDS by feature cov mean. add_feature_cov_mean() must be called before this function.
 
@@ -242,6 +326,8 @@ class MCDS(xr.Dataset):
         -------
         MCDS
         """
+        var_dim = self._verify_dim(var_dim, mode='var')
+
         try:
             feature_cov_mean = self.coords[f"{var_dim}_cov_mean"].to_pandas()
         except KeyError:
@@ -259,7 +345,7 @@ class MCDS(xr.Dataset):
         print(f" After cov mean filter: {selected_feature} {var_dim} {ratio:.1f}%")
         return mcds
 
-    def get_feature_bed(self, var_dim):
+    def get_feature_bed(self, var_dim=None):
         """
         Get a bed format data frame of the var_dim
 
@@ -271,6 +357,7 @@ class MCDS(xr.Dataset):
         -------
         pd.DataFrame
         """
+        var_dim = self._verify_dim(var_dim, mode='var')
 
         bed_df = pd.DataFrame(
             [
@@ -283,7 +370,7 @@ class MCDS(xr.Dataset):
         ).T
         return bed_df
 
-    def remove_black_list_region(self, var_dim, black_list_path, f=0.2):
+    def remove_black_list_region(self, black_list_path, var_dim=None, f=0.2):
         """
         Remove regions overlap (bedtools intersect -f {f}) with regions in the black_list_path
 
@@ -299,6 +386,8 @@ class MCDS(xr.Dataset):
         -------
         MCDS
         """
+        var_dim = self._verify_dim(var_dim, mode='var')
+
         with warnings.catch_warnings():
             warnings.simplefilter("ignore")
             feature_bed_df = self.get_feature_bed(var_dim=var_dim)
@@ -308,16 +397,16 @@ class MCDS(xr.Dataset):
             try:
                 black_feature_index = (
                     black_feature.to_dataframe()
-                    .set_index(["chrom", "start", "end"])
-                    .index
+                        .set_index(["chrom", "start", "end"])
+                        .index
                 )
             except pd.errors.EmptyDataError:
                 print("No feature overlapping the black list bed file.")
                 return self
         black_feature_id = pd.Index(
             feature_bed_df.reset_index()
-            .set_index(["chrom", "start", "end"])
-            .loc[black_feature_index][var_dim]
+                .set_index(["chrom", "start", "end"])
+                .loc[black_feature_index][var_dim]
         )
 
         print(
@@ -328,7 +417,7 @@ class MCDS(xr.Dataset):
             mcds = self.sel({var_dim: ~self.get_index(var_dim).isin(black_feature_id)})
         return mcds
 
-    def remove_chromosome(self, var_dim, exclude_chromosome):
+    def remove_chromosome(self, exclude_chromosome, var_dim=None):
         """
         Remove regions in specific chromosome
 
@@ -342,6 +431,7 @@ class MCDS(xr.Dataset):
         -------
         MCDS (xr.Dataset)
         """
+        var_dim = self._verify_dim(var_dim, mode='var')
         judge = self.coords[f"{var_dim}_chrom"].isin(exclude_chromosome)
         print(f"{int(judge.sum())} {var_dim} features in {exclude_chromosome} removed.")
         with dask.config.set(**{"array.slicing.split_large_chunks": False}):
@@ -349,35 +439,36 @@ class MCDS(xr.Dataset):
         return mcds
 
     def calculate_hvf_svr(
-        self,
-        var_dim,
-        mc_type,
-        obs_dim="cell",
-        n_top_feature=5000,
-        da_suffix="frac",
-        plot=True,
+            self,
+            mc_type=None,
+            var_dim=None,
+            obs_dim=None,
+            n_top_feature=5000,
+            da_suffix="frac",
+            plot=True,
     ):
         from sklearn.svm import SVR
         import plotly.graph_objects as go
+        obs_dim = self._verify_dim(obs_dim, mode='obs')
+        var_dim = self._verify_dim(var_dim, mode='var')
 
-        feature_mc_frac_mean = (
-            self[f"{var_dim}_da_{da_suffix}"]
-            .sel(mc_type=mc_type)
-            .mean(dim=obs_dim)
-            .to_pandas()
-        )
-        feature_std = (
-            self[f"{var_dim}_da_{da_suffix}"]
-            .sel(mc_type=mc_type)
-            .std(dim=obs_dim)
-            .to_pandas()
-        )
-        feature_cov_mean = (
-            self[f"{var_dim}_da"]
-            .sel(mc_type=mc_type, count_type="cov")
-            .mean(dim=obs_dim)
-            .to_pandas()
-        )
+        frac_da = self[f"{var_dim}_da_{da_suffix}"]
+        count_da = self[f"{var_dim}_da"]
+        if 'mc_type' in self.dims:
+            if mc_type is None:
+                mc_types = self.get_index('mc_type')
+                if mc_types.size > 1:
+                    raise ValueError(f'Data array has multiple mc_types, '
+                                     f'please specify mc_type parameter.')
+                else:
+                    mc_type = mc_types[0]
+            feature_mc_frac_mean = frac_da.sel(mc_type=mc_type).mean(dim=obs_dim).to_pandas()
+            feature_std = frac_da.sel(mc_type=mc_type).std(dim=obs_dim).to_pandas()
+            feature_cov_mean = count_da.sel(mc_type=mc_type, count_type="cov").mean(dim=obs_dim).to_pandas()
+        else:
+            feature_mc_frac_mean = frac_da.mean(dim=obs_dim).to_pandas()
+            feature_std = frac_da.std(dim=obs_dim).to_pandas()
+            feature_cov_mean = count_da.sel(count_type="cov").mean(dim=obs_dim).to_pandas()
 
         # remove bad features
         judge = (feature_mc_frac_mean > 0) & (feature_std > 0) & (feature_cov_mean > 0)
@@ -388,7 +479,7 @@ class MCDS(xr.Dataset):
             )
         feature_mc_frac_mean = feature_mc_frac_mean[judge]
         feature_var = (
-            feature_std[judge] ** 2
+                feature_std[judge] ** 2
         )  # to be consistent with previous bin-based method, use var here
         feature_cov_mean = feature_cov_mean[judge]
 
@@ -444,8 +535,8 @@ class MCDS(xr.Dataset):
                         marker=dict(
                             size=2,
                             color=plot_data["feature_select"]
-                            .map({True: "red", False: "gray"})
-                            .tolist(),  # set color to an array/list of desired values
+                                .map({True: "red", False: "gray"})
+                                .tolist(),  # set color to an array/list of desired values
                             opacity=0.8,
                         ),
                     )
@@ -462,23 +553,27 @@ class MCDS(xr.Dataset):
             fig.show()
 
         for name, column in hvf_df.items():
-            self.coords[f"{var_dim}_{mc_type}_{name}"] = column
+            if mc_type is None:
+                self.coords[f"{var_dim}_{name}"] = column
+            else:
+                self.coords[f"{var_dim}_{mc_type}_{name}"] = column
         return hvf_df
 
     def calculate_hvf(
-        self,
-        mc_type,
-        var_dim,
-        obs_dim="cell",
-        min_disp=0.5,
-        max_disp=None,
-        min_mean=0,
-        max_mean=5,
-        n_top_feature=5000,
-        bin_min_features=5,
-        mean_binsize=0.05,
-        cov_binsize=100,
-        plot=True,
+            self,
+            mc_type=None,
+            var_dim=None,
+            obs_dim=None,
+            min_disp=0.5,
+            max_disp=None,
+            min_mean=0,
+            max_mean=5,
+            n_top_feature=5000,
+            bin_min_features=5,
+            mean_binsize=0.05,
+            cov_binsize=100,
+            da_suffix="frac",
+            plot=True,
     ):
         """
         Calculate normalized dispersion to select highly variable features.
@@ -516,7 +611,22 @@ class MCDS(xr.Dataset):
         -------
         pd.DataFrame
         """
-        matrix = self[f"{var_dim}_da_frac"].sel(mc_type=mc_type).squeeze()
+        obs_dim = self._verify_dim(obs_dim, mode='obs')
+        var_dim = self._verify_dim(var_dim, mode='var')
+
+        frac_da = self[f"{var_dim}_da_{da_suffix}"]
+        if 'mc_type' in self.dims:
+            if mc_type is None:
+                mc_types = self.get_index('mc_type')
+                if mc_types.size > 1:
+                    raise ValueError(f'Data array has multiple mc_types, '
+                                     f'please specify mc_type parameter.')
+                else:
+                    mc_type = mc_types[0]
+            matrix = frac_da.sel(mc_type=mc_type).squeeze()
+        else:
+            matrix = frac_da.squeeze()
+
         feature_mean_cov = self.coords[f"{var_dim}_cov_mean"]
 
         hvf_df = highly_variable_methylation_feature(
@@ -551,10 +661,16 @@ class MCDS(xr.Dataset):
             )
 
         for name, column in hvf_df.items():
-            self.coords[f"{var_dim}_{mc_type}_{name}"] = column
+            if mc_type is None:
+                self.coords[f"{var_dim}_{name}"] = column
+            else:
+                self.coords[f"{var_dim}_{mc_type}_{name}"] = column
         return hvf_df
 
-    def get_score_adata(self, var_dim, mc_type, quant_type, obs_dim='cell', sparse=True):
+    def get_score_adata(self, mc_type, quant_type, obs_dim=None, var_dim=None, sparse=True):
+        obs_dim = self._verify_dim(obs_dim, mode='obs')
+        var_dim = self._verify_dim(var_dim, mode='var')
+
         da = self[f'{var_dim}_da_{mc_type}-{quant_type}']
         if sparse:
             from scipy.sparse import csr_matrix
@@ -567,13 +683,13 @@ class MCDS(xr.Dataset):
         return adata
 
     def get_adata(
-        self,
-        mc_type,
-        var_dim,
-        da_suffix="frac",
-        obs_dim="cell",
-        select_hvf=True,
-        split_large_chunks=True,
+            self,
+            mc_type=None,
+            obs_dim=None,
+            var_dim=None,
+            da_suffix="frac",
+            select_hvf=True,
+            split_large_chunks=True,
     ):
         """
         Get anndata from MCDS mC rate matrix
@@ -596,39 +712,62 @@ class MCDS(xr.Dataset):
         -------
         anndata.Anndata
         """
+        obs_dim = self._verify_dim(obs_dim, mode='obs')
+        var_dim = self._verify_dim(var_dim, mode='var')
+
+        if 'mc_type' in self.dims:
+            if mc_type is None:
+                mc_types = self.get_index('mc_type')
+                if mc_types.size > 1:
+                    raise ValueError(f'Data array has multiple mc_types, '
+                                     f'please specify mc_type parameter.')
+                else:
+                    mc_type = mc_types[0]
+        else:
+            mc_type = None
+
         with dask.config.set(
-            **{"array.slicing.split_large_chunks": split_large_chunks}
+                **{"array.slicing.split_large_chunks": split_large_chunks}
         ):
-            if select_hvf:
-                try:
-                    use_features = (
-                        self.coords[f"{var_dim}_{mc_type}_feature_select"]
-                        .to_pandas()
-                        .dropna()
-                        .astype(bool)
-                    )
-                    use_features = use_features[use_features].index
-                    use_data = (
-                        self[f"{var_dim}_da_{da_suffix}"]
-                        .sel({"mc_type": mc_type, var_dim: use_features})
-                        .squeeze()
-                    )
-                except KeyError:
-                    print(
-                        "select_hvf is True, but no highly variable feature results found, "
-                        "use all features instead."
-                    )
-                    use_data = (
-                        self[f"{var_dim}_da_{da_suffix}"]
-                        .sel({"mc_type": mc_type})
-                        .squeeze()
-                    )
+            frac_da = self[f"{var_dim}_da_{da_suffix}"]
+            if mc_type is None:
+                if select_hvf:
+                    try:
+                        use_features = (
+                            self.coords[f"{var_dim}_feature_select"]
+                                .to_pandas()
+                                .dropna()
+                                .astype(bool)
+                        )
+                        use_features = use_features[use_features].index
+                        use_data = frac_da.sel({var_dim: use_features}).squeeze()
+                    except KeyError:
+                        print(
+                            "select_hvf is True, but no highly variable feature results found, "
+                            "use all features instead."
+                        )
+                        use_data = frac_da.squeeze()
+                else:
+                    use_data = frac_da.squeeze()
             else:
-                use_data = (
-                    self[f"{var_dim}_da_{da_suffix}"]
-                    .sel({"mc_type": mc_type})
-                    .squeeze()
-                )
+                if select_hvf:
+                    try:
+                        use_features = (
+                            self.coords[f"{var_dim}_{mc_type}_feature_select"]
+                                .to_pandas()
+                                .dropna()
+                                .astype(bool)
+                        )
+                        use_features = use_features[use_features].index
+                        use_data = frac_da.sel({"mc_type": mc_type, var_dim: use_features}).squeeze()
+                    except KeyError:
+                        print(
+                            "select_hvf is True, but no highly variable feature results found, "
+                            "use all features instead."
+                        )
+                        use_data = frac_da.sel({"mc_type": mc_type}).squeeze()
+                else:
+                    use_data = frac_da.sel({"mc_type": mc_type}).squeeze()
 
             obs_df = pd.DataFrame([], index=use_data.get_index(obs_dim).astype(str))
             var_df = pd.DataFrame([], index=use_data.get_index(var_dim).astype(str))
@@ -660,13 +799,15 @@ class MCDS(xr.Dataset):
         return adata
 
     def merge_cluster(
-        self,
-        cluster_col,
-        obs_dim="cell",
-        add_mc_frac=True,
-        add_overall_mc=True,
-        overall_mc_da="chrom100k_da",
+            self,
+            cluster_col,
+            obs_dim=None,
+            add_mc_frac=True,
+            add_overall_mc=True,
+            overall_mc_da="chrom100k_da",
     ):
+        obs_dim = self._verify_dim(obs_dim, mode='obs')
+
         if isinstance(cluster_col, str):
             cluster_col = cluster_col
         else:
@@ -683,30 +824,138 @@ class MCDS(xr.Dataset):
                         cluster_mcds.add_mc_frac(var_dim=name[:-3])
 
             if add_overall_mc:
-                for mc_type in cluster_mcds.get_index("mc_type"):
+                if 'mc_type' in cluster_mcds.dims:
+                    for mc_type in cluster_mcds.get_index("mc_type"):
+                        overall_mc_dim = overall_mc_da[:-3]
+                        mc = (
+                            cluster_mcds[overall_mc_da]
+                                .sel(mc_type=mc_type, count_type="mc")
+                                .sum(dim=overall_mc_dim)
+                        )
+                        cov = (
+                            cluster_mcds[overall_mc_da]
+                                .sel(mc_type=mc_type, count_type="cov")
+                                .sum(dim=overall_mc_dim)
+                        )
+                        cluster_mcds.coords[f"{cluster_col}_{mc_type}_overall"] = mc / cov
+                else:
                     overall_mc_dim = overall_mc_da[:-3]
-                    mc = (
-                        cluster_mcds[overall_mc_da]
-                        .sel(mc_type=mc_type, count_type="mc")
-                        .sum(dim=overall_mc_dim)
-                    )
-                    cov = (
-                        cluster_mcds[overall_mc_da]
-                        .sel(mc_type=mc_type, count_type="cov")
-                        .sum(dim=overall_mc_dim)
-                    )
-                    cluster_mcds.coords[f"{cluster_col}_{mc_type}"] = (
-                        mc / cov
-                    ).to_pandas()
+                    mc = cluster_mcds[overall_mc_da].sel(count_type="mc").sum(dim=overall_mc_dim)
+
+                    cov = cluster_mcds[overall_mc_da].sel(count_type="cov").sum(dim=overall_mc_dim)
+                    cluster_mcds.coords[f"{cluster_col}_overall"] = mc / cov
         return cluster_mcds
 
-    def to_region_ds(self, region_dim):
-        from .region_ds import RegionDS
+    def to_region_ds(self, region_dim=None):
+        region_dim = self._verify_dim(region_dim, mode='var')
 
+        from .region_ds import RegionDS
         _region_ds = RegionDS(self, region_dim=region_dim)
         return _region_ds
 
-    # TODO
-    # def save()
-    # save mcds to a location, if the location is a MCDS dataset, auto save to sub_dir and modify .ALLCools
-    # deal with large save issue, load by chunk and write append
+    def write_dataset(self,
+                      output_path,
+                      mode='w-',
+                      obs_dim=None,
+                      var_dims: Union[str, list] = None,
+                      use_obs=None,
+                      chunk_size=1000):
+        """
+        Write MCDS into a on-disk zarr dataset. Data arrays for each var_dim will be saved in separate
+        sub-directories of output_path.
+
+        Parameters
+        ----------
+        output_path
+            Path of the zarr dataset
+        mode
+            'w-' means write to output_path, fail if the path exists; 'w' means write to output_path,
+            overwrite if the var_dim sub-directory exists
+        obs_dim
+            dimension name of observations
+        var_dims
+            dimension name, or a list of dimension names of variables
+        use_obs
+            Select AND order observations when write.
+        chunk_size
+            The load and write chunks, set this as large as possible based on available memory.
+
+        Returns
+        -------
+        output_path
+        """
+        # save mcds to a location, if the location is a MCDS dataset, auto save to sub_dir and modify .ALLCools
+        # deal with large save issue, load by chunk and write append
+        obs_dim = self._verify_dim(obs_dim, mode='obs')
+        if var_dims is None:
+            var_dims = [self._verify_dim(None, mode='var')]
+        else:
+            if isinstance(var_dims, str):
+                var_dims = [var_dims]
+            else:
+                pass
+
+        # determine output_path type
+        output_path = pathlib.Path(output_path)
+        config_path = pathlib.Path(f'{output_path}/.ALLCools')
+        if config_path.exists():
+            if mode == 'w-':
+                raise FileExistsError(f'{output_path} already exist. Use mode="a" '
+                                      f'if you want to add additional data array to exist dataset. '
+                                      f'Or use mode="w" to overwrite existing dataset')
+            else:
+                import shutil
+                shutil.rmtree(output_path)
+
+        # separate data_vars by var_dim
+        data_vars_dict = {k: [] for k in var_dims}
+        for da_name, da in self.data_vars.items():
+            var_dim_count = 0
+            for var_dim in var_dims:
+                if var_dim in da.dims:
+                    var_dim_count += 1
+                    if var_dim_count == 1:
+                        data_vars_dict[var_dim].append(da_name)
+                    else:
+                        raise ValueError(f'{da_name} contain more then one var_dim. '
+                                         f'Dims of {da_name} is {da.dims}')
+
+        obs_index = self.get_index(obs_dim)
+        for var_dim, da_names in data_vars_dict.items():
+            _this_output_path = f'{output_path}/{var_dim}'
+            if pathlib.Path(_this_output_path).exists():
+                print(f'{output_path}/{var_dim} will be overwrote by new data array')
+
+            if len(da_names) == 0:
+                print(f'No data for var_dim {var_dim}.')
+            else:
+                _mcds = self[da_names]
+                # pop out obs_dim and var_dim from attrs, do not save these to dataset
+                _mcds.attrs.pop('obs_dim')
+                _mcds.attrs.pop('var_dim')
+
+                for i, chunk_start in enumerate(range(0, obs_index.size, chunk_size)):
+                    if use_obs is None:
+                        _obs_chunk = obs_index[chunk_start:chunk_start + chunk_size]
+                        _chunk_ds = _mcds.sel({obs_dim: _obs_chunk}).load()
+                    else:
+                        _obs_chunk = use_obs[chunk_start:chunk_start + chunk_size]
+                        # load first, order next, to prevent unordered index causing more chunks
+                        _chunk_ds = _mcds.sel({obs_dim: obs_index.isin(_obs_chunk)}).load()
+                        _chunk_ds = _chunk_ds.sel({obs_dim: _obs_chunk})
+
+                    if i == 0:
+                        _chunk_ds.to_zarr(_this_output_path, mode='w')
+                    else:
+                        _chunk_ds.to_zarr(_this_output_path, append_dim=obs_dim)
+
+        # write config
+        from .utilities import update_dataset_config
+        update_dataset_config(output_path,
+                              add_ds_region_dim={var_dim: var_dim
+                                                 for var_dim, da_list in data_vars_dict.items()
+                                                 if len(da_list) > 0},
+                              add_ds_sample_dim={var_dim: obs_dim
+                                                 for var_dim, da_list in data_vars_dict.items()
+                                                 if len(da_list) > 0})
+        return output_path
